package com.portlandwebworks.commons.dao;

import com.portlandwebworks.commons.domain.EntityBase;
import com.portlandwebworks.commons.domain.Identifiable;
import org.apache.poi.hssf.record.formula.functions.T;

import java.io.Serializable;
import java.util.Collection;
import java.util.List;
import java.util.Map;

/**
 * This software is the property of Portland Webworks.
 *
 * @author      Ronnie Northup
 */
public interface IEntityManager
{
	/**
	 * A last resort API.  This will return the underlying implementation-specific context for peforming database
	 * operations not supported by this interface.
	 *
	 * @return
	 */
	public Object getPersistenceContext();

	/**
	 * Allows assigning a custom subclass of <code>PagedSearchResults</code> to this entity manager.  All methods that
	 * return an instance of <code>PagedSearchResult</code> will return this type instead.
	 *
	 * @param pagedSearchResultType     The custom subclass this entity manager should return to the caller.
	 */
	void setPagedSearchResultType(Class<PagedSearchResult> pagedSearchResultType);

	/**
	 * Loads an object based on its unique identifier.  This object can be a persisted entity, or it can be a POJO that,
	 * through @EntityMapping and @FieldMapping annotations, is "mapped" to a class of persisted entity.
	 *
	 * @param returnType    The type of entity or entity-mapped POJO to be retrieved.
	 * @param id            The identifier of the desired entity
	 * @param fetchOptions  The agregate fields on the entity to be eager loaded, if possible, or to be retrieved after
	 *                      the root entity query if the data can't be joined to the root query.
	 *
	 * @return              The entity or POJO, populated with persisted data.
	 *
	 * @throws DataNotFoundException        If an entity with the specified identifier could not be found.
	 * @throws IllegalArgumentException     If the return type is not a valid return type (is not a known entity or does
	 *                                      not map to a known entity), or if the identifier provided is not a valid
	 *                                      identifier for the target or mapped entity.
	 */
	public <T> T findById(Class<T> returnType, Serializable id, Object... fetchOptions) throws DataNotFoundException, IllegalArgumentException;

	/**
	 *
	 * @return  @param entityClass  @param fetchOptions
	 */
	public <T> List<T> findAll(Class<T> returnType, Object... fetchOptions);

	/**
	 * Returns a mapping of entities
	 *
	 * @param returnType
	 * @param keyField
	 * @param fetchOptions
	 * @return
	 */
	public <T, K extends Serializable> Map<K, T> mapAll(Class<T> returnType, String keyField, Object... fetchOptions);

	/**
	 * Retrieves a collection of entities by their primary key values.
	 *
	 * @param returnType
	 * @param ids
	 * @param fetchOptions
	 * @return
	 * @throws DataNotFoundException
	 */
	public <T> List<T> findAllByIds(Class<T> returnType, Serializable[] ids, Object... fetchOptions) throws DataNotFoundException;

	/**
	 * Retrieves a list of entities or POJOs that match the criteria provided, with zero or more aggregate fields eager loaded.
	 *
	 * @param returnType   The type of object to be retrieved.
	 * @param criteria      The criteria to apply.
	 * @param fetchOptions  The fields to be eager loaded.
	 *
	 * @return              A list of entities or POJOs with an entity mapping, depending upon the
	 */
	public <T> List<T> findAllByCriteria(Class<T> returnType, SearchCriteria criteria, Object... fetchOptions);
	
	/**
	 * Rudimentary API to retrieve a list of entities filtered using one or more predicates ANDed together.
	 *
	 * @param returnType
	 * @param predicates
	 * @param <T>
	 * @return
	 */
	public <T> List<T> findAllByPredicate(Class<T> returnType, Predicate... predicates);

	/**
	 * When you just want one field of an entity.
	 * 
	 * @param rootEntity
	 * @param field
	 * @param predicates
	 * @param <T>
	 * @return
	 */
	public <T> T projectByPredicate(Class<? extends Identifiable> rootEntity, String field, Predicate... predicates);

	/**
	 * When you just want one field from a set of entities matching some predicates.
	 * 
	 * @param rootEntity
	 * @param field
	 * @param predicates
	 * @param <T>
	 * @return
	 */
	public <T> List<T> projectAllByPredicate(Class<? extends Identifiable> rootEntity, String field, Predicate... predicates);

	/**
	 *
	 * @param rootEntity
	 * @param field
	 * @param sortSpec
	 * @param predicates
	 * @param <T>
	 * @return
	 */
	public <T> List<T> projectAllByPredicate(Class<? extends Identifiable> rootEntity, String field, SortSpecification sortSpec, Predicate... predicates);

	/**
	 * For when you just want a subset of fields from an entity matching one or more predicates.
	 *
	 * @param rootEntity
	 * @param fields
	 * @param predicates
	 * @return
	 */
	public Map projectByPredicate(Class<? extends Identifiable> rootEntity, String[] fields, Predicate... predicates);
	
	/**
	 * For when you want a subset of fields from an entity or its associated entities matching one or more predicates.
	 *
	 * @param rootEntity
	 * @param fields
	 * @param predicates
	 * @return
	 */
	public List<Map> projectAllByPredicate(Class<? extends Identifiable> rootEntity, String[] fields, Predicate... predicates);

	/**
	 *
	 * @param rootEntity
	 * @param fields
	 * @param sortSpec
	 * @param predicates
	 * @return
	 */
	public List<Map> projectAllByPredicate(Class<? extends Identifiable> rootEntity, String[] fields, SortSpecification sortSpec, Predicate... predicates);

	/**
	 * Retrieves a list of entities or pojos sorted in a particular order
	 * 
	 * @param returnType
	 * @param sort
	 * @param predicates
	 * @param <T>
	 * @return
	 */
	public <T> List<T> findAllByPredicate(Class<T> returnType, SortSpecification sort, Predicate... predicates);

	/**
	 * Retrievs a single entity filtered using one or more predicates.
	 *
	 * @param returnType
	 * @param predicates
	 *
	 * @return
	 */
	public <T> T findByPredicate(Class<T> returnType, Predicate... predicates);

	/**
	 * Generic method to return a "page" of search results suitable for display in a grid view.
	 *
	 * @param returnType
	 * @param criteria
	 * @param options
	 * @param fetchOptions
	 * @return
	 */
	public <T> PagedSearchResult<T> search(Class<T> returnType, SearchCriteria criteria, IPagedSearchOptions options, Object... fetchOptions);


	/**
	 * Generic method to return a page of search results suitable for display in a grid view.  This specific
	 * API allows search criteria to be applied to one class, and to return as search results another class.  This
	 * allows one to run a complex set of query criteria against a domain entity, and to return as paged results
	 * a "flattened" view version of the entity.
	 *
	 * @param viewClass         The type of each return value returned by the search method.  This is useful for
	 *                          generating "flattened" results of a query against an entity type that has a lot of
	 *                          aggregate data, or for scenarios where the rendered view requires complex calculations
	 *                          on the queried entities.  The unique ID of this class must correspond to the unique ID
	 *                          of the criteria class.
	 *
	 * @param criteriaClass     The entity which is queried, that is, to which the search criteria will be applied.  The
	 *                          unique ID of this class must correspond to the unique ID of the view class.
	 *
	 * @param searchCriteria
	 * @param options
	 * @param fetchOptions
	 * @return
	 * @deprecated              This method uses an obsolete method for returning a "view" representation of entity
	 *                          search results by providing an Identifiable as the view class which is backed by a database
	 *                          view.  The new way of performing a search return view data is to call
	 *                          #search(Class<T> entityClass, SearchCriteria criteria, PagedSearchOptions options, Enum... fetchOptions)
	 */
	@Deprecated
	public <T extends EntityBase, Y extends EntityBase> PagedSearchResult<T> search(Class<T> viewClass, Class<Y> criteriaClass, SearchCriteria searchCriteria, IPagedSearchOptions options, Object... fetchOptions);

	/**
	 * Inserts a new instance of entity into the data source.
	 *
	 * @param entity
	 * @return
	 * @throws DuplicateDataException
	 */
	public <T extends Identifiable> Serializable insert(T entity) throws DuplicateDataException;

	/**
	 * Added because Jacob doesn't like writing code.
	 *
	 * @param entity
	 * @param <T>
	 * @return
	 * @throws DuplicateDataException
	 */
	public <T extends Identifiable> void saveOrUpdate(T entity) throws DuplicateDataException;

	/**
	 * Propagates local entity changes to the underlying database.  The actual propagation may not happen until the end
	 * of the transaction, depending upon how the ORM implementation behaves.
	 *
	 * @param entity
	 * @throws DuplicateDataException
	 */
	public <T extends Identifiable> void update(T entity) throws DuplicateDataException;

	/**
	 * Removes an entity from the data source.
	 *
	 * @param entity
	 */
	public <T extends Identifiable> void delete(T entity) throws DataIntegrityException, IllegalArgumentException, DataNotFoundException;

	/**
	 * Removes an entity from the data source that has not been loaded by referencing it by its identifier.
	 *
	 * @param entityClass               The entity type to delete.
	 * @param id                        The entity ID to delete.
	 * @throws DataNotFoundException    If there is no persisted entity of the type specified with the given ID.
	 */
	public <T extends Identifiable> void deleteById(Class<T> entityClass, Serializable id) throws DataNotFoundException;

	/**
	 *
	 *
	 * @param entityClass
	 * @param ids
	 * @throws DataNotFoundException
	 */
	<T extends Identifiable> void deleteAllById(Class<T> entityClass, Collection<? extends Serializable> ids) throws DataNotFoundException;

	/**
	 *
	 * @param entityClass
	 * @param ids
	 * @param <T>
	 * @throws DataNotFoundException
	 */
	public <T extends Identifiable> void deleteAllById(Class<T> entityClass, Serializable... ids) throws DataNotFoundException;

	/**
	 * Removes a collection of entities from the data source by their unique identifier.
	 *
	 * Note--this method of deletion does not take into account any cascade options configured at the entity level.
	 *
	 * @param entities
	 */
	public <T extends Identifiable> void deleteAll(Collection<T> entities);

	/**
	 * This should force the DAO implementation to flush all pending changes to the database without yet committing.
	 */
	public void saveChanges();

	/**
	 * Executes a text query in the underlying persistence query language, returning a list of entities of the specified
	 * type.  The type to be returned in the query is determined by the query itself.
	 *
	 * @param returnType    The desired type for packaging the retrieved data.  Types will be handled in the following way:
	 *                      <ul>
	 *                          <li>
	 *                              Map - Each item in the returned list is a Map, whose keys are the selected column
	 *                              names and values are the selected column values.  NOTE--The map keys are determined
	 *                              by parsing aliases selected in the query.  You must specify aliases in your query.
	 *                              If you want to specify types for the values stored in the map, or if you want to
	 *                              return nested entities as part of your query results, use #executeQuery(Class, String, Object..)
	 *                              instead, which allows explicitly declaring the types of each column value in the SQL
	 *                              results.
	 *                          </li>
	 *                          <li>
	 *                              List - each item in the returned list is a List of each row of selected values.
	 *                          </li>
	 *                          <li>
	 *                              Array - default Hibernate behavior for multiple return values.  Each item in the
	 *                              returned list is an Object[] array of each row of selected values.
	 *                          </li>
	 *                          <li>
	 *                              Identifiable - Each item in the returned list is an instance of the Identifiable
	 *                              subclass specified, whose members are populated with a row of query results.
	 *                          </li>
	 *                          <li>
	 *                              Value Type - See FieldUtils.getFieldType for the definition of what is a value type.
	 *                              Each item in the returned list is an instance
	 *                          </li>
	 *                      </ul>
	 * @param query         The query text used to execute the query, using the language of the configured persistence
	 *                      implementation.
	 * @param params        The parameters to apply to the query.  The following types of data are accepted as
	 *                      query parameters:
	 *                      <ul>
	 *                          <li>
	 *                          An object that exposes java bean properties that can be used to supply values for the
	 *                          named query parameters in the query provided.  The property names must map to the query
	 *                          parameters.  If the query contains a named parameter that does not have a bean property
	 *                          of the same name, an exception is thrown.
     *                          </li>
	 *                          <li>
	 *                          A Map implementation whose key/value pairs can be used to supply values for the named
	 *                          query parameters in the query provided.
     *                          </li>
	 *                          <li>
	 *                          A list of vararg values that correspond in order to the ? style query parameters i
	 *                          the query provided.  This parameter option is only available if the query uses ?
	 *                          placeholders instead of named placeholders.
	 *                          </li>
	 *                      </ul>
	 * @return              A <code>List</code> of values whose type varies based on the <code>returnType</code> specified.
	 */
	public <T> List<T> executeQuery(Class<T> returnType, String query, Object... params);

	/**
	 * Executes a query that updates records in the database.
	 *
	 * @param hql       The implementation-specific query to execute.
	 * @param params    The parameters to apply to the update.  See #executeQuery for a list of accepted values.
	 * @return          The total number of objects affected by the update.
	 */
	public int executeUpdate(String hql, Object... params);

	/**
	 * Executes a "named" query using the underlying persistence's query configuration mechanism, returning a list of
	 * entities of the specified type.
	 *
	 * @param name          The configured name of the query.
	 * @param params        The parameters to apply to the query.  See #executeQuery for a list of accepted values.
	 *
	 * @return
	 */
	public <T> List<T> executeNamedQuery(String name, Object... params);

	/**
	 * Executes an SQL statement that updates or deletes records in the database.
	 *
	 * @param sql           The SQL, preferably ANSI-92 compliant
	 * @param params        The parameters to apply to the SQL.  See #executeQuery for a list of accepted values.
	 * @return              The total number of records affected by the SQL statement execution.
	 */
	public int executeSQLUpdate(String sql, Object... params);

	/**
	 * Executes a raw SQL query, returning a list of entities of the specific type.  Highly recommended that you adhere
	 * to the ANSI-92 SQL standard to allow your code to work with different persistence implementations.
	 *
	 * If the SQL query has
	 *
	 * @param returnType
	 * @param sql           The SQL query to execute
	 * @param params        The parameters to apply to the SQL.  See #executeQuery for a list of accepted values.
	 * @return              A <code>List</code> of values whose type varies based on the <code>returnType</code> specified.
	 */
	public <T> List<T> executeSQLQuery(Class<T> returnType, String sql, Object... params);

	/**
	 * Executes a raw SQL query, returning a map of aliased values of the types specified.  Best if you supply ANSI-92
	 * SQL standards to your query.
	 *
	 * @param returnTypes
	 * @param sql
	 * @param params        The parameters to apply to the SQL.  See #executeQuery for a list of accepted values.
	 * @return
	 */
	public List<Map<String, Class<?>>> executeSQLQuery(Map<String, Class<?>> returnTypes, String sql, Object... params);

	/**
	 * Executes a raw SQL query stored in a configuration file that has been loaded by the persistence engine.  The SQL
	 * query's configured return value is assumed to be as set of scalar values that will be converted to the specified
	 * return type.  This method does not rely on the configuration of the SQL query to determine what the class
	 * representation of the return value is.
	 * </p>
	 * <p>
	 * If the return type is an entity type, the entity will be populated from the raw result set values.  The aliases
	 * of the result set data must match the entity fields.
	 * <br/>
	 * If the return type is Map, each result set row will become a Map instance, where the column aliases are the keys
	 * that map to the result set data.
	 * <br/>
	 * If the return type is List, each result set row will become a List instance, where each column value is indexed
	 * correspondingly in the list.
	 * <br/>
	 * If Object[] is specified as the return type, each result set row will be returned as an Object array.
	 * <br/>
	 * If any primitive type or wrapper of a primitive type, or a String or Date type is specified as the return type,
	 * the results of the query will be returned as a List of said types.
	 * <br/>
	 * If the return type is not that of a persisted entity, each non-transient field in the class is expected to map
	 * to an alias in the result set, and the field's type should correspond to the type of the result set data.
	 * </p>
	 *
	 * @param returnType    The desired return type in which to encapsulate scalar result set data.
	 * @param name          The name of the configured query to execute.
	 * @param params        The parameters to apply to the SQL.  See #executeQuery for a list of accepted values.
	 *
	 * @return
	 */
	public <T> List<T> executeNamedSQLQuery(Class<T> returnType, String name, Object... params);

	/**
	 * <p>
	 *  Executes a query for persisted entities using a named SQL script stored using the underlying ORM's mechanism.
	 *  The entity type to be returned to the caller is controlled by the configuration of the SQL script, instead of
	 *  passing the type as an argument to this method.
	 * </p>
	 * <p>
	 *  This method is only supported for ORMs that allow the configuration of the SQL query to indicate what the return
	 *  value(s) should be.  The caller of this method is expected to know what the return values will be for the query.
	 * </p>
	 *
	 *
	 * @param name          The name of the query to execute.
	 * @param params        The parameters to apply to the SQL.  See #executeQuery for a list of accepted values.
	 *
	 * @return
	 */
	public <T> List<T> executeNamedEntitySQL(String name, Object... params);

	/**
	 * Some persistence frameworks cache entities in memory to optimize repeated access to the same entities during a
	 * session.  This method removes an entity from the memory cache.  This allows changes to the entity that will not
	 * be propagated to the database when the session's transaction is committed, or saveChanges() is called.
	 *
	 * @param identifiable
	 * @param <T>
	 */
	<T extends Serializable> void removeFromContext(Identifiable<T> identifiable);

	/**
	 * Gets a count of all entities who meet the specified criteria.
	 * 
	 *
	 *
	 * @param entityClass
	 * @param searchCriteria
	 * @return
	 */
	long getCount(Class<?> entityClass, SearchCriteria searchCriteria);
}
