/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.dao;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;


/**
 * This software is the property of Portland Webworks.
 *
 * Add this annotation to the fields in your SearchCriteria class to tell a DAO implementation how
 * to apply those search criteria to a search query.
 *
 * @Author rnorthup
 */
@Retention(RetentionPolicy.RUNTIME)
@Target({ElementType.FIELD, ElementType.METHOD, ElementType.ANNOTATION_TYPE, ElementType.PARAMETER})
public @interface SearchCriterion
{
	/**
	 * The type of query operator to apply.  Indicates how this search criterion should be applied to the target
	 * domain object that is being queried.
	 *
	 * If no operator is specified, this criterion defaults to Operator.EqualTo
	 *
	 * @See com.pww.commons.dao.Operator
	 *
	 * @return
	 */
	public abstract Operator operator() default Operator.EqualTo;

	/**
	 * The name of a property on the target entity (or one of its aggregate entities, specified using the joinPath
	 * field) to which the operator will be applied for this search operation.  If unspecified, the name of the
	 * property defaults to the name of the field to which this annotation is applied.
	 *
	 * @return
	 */
	public abstract String propertyName() default "";

	/**
	 * Indicates whether an empty string value for this search criteria should be applied as part of the query.  By
	 * default, an empty string value will be included in the database query, testing the target property for an empty
	 * value.  Setting this property to false will exclude this criteria from the query if an empty string value
	 * is provided.
	 * 
	 * @return
	 */
	public abstract boolean useEmptyStringValue() default true;

	/**
	 * Specifying a value for this criterion overrides the value stored in the field to which the annotation is applied.
	 * 
	 * @return
	 */
	public String value() default "";
}