package com.portlandwebworks.commons.data;

import com.portlandwebworks.commons.collections.ListUtils;
import com.portlandwebworks.commons.dao.DaoUtils;
import com.portlandwebworks.commons.dao.DataNotFoundException;
import com.portlandwebworks.commons.dao.IEntityManager;
import com.portlandwebworks.commons.dao.Predicate;
import com.portlandwebworks.commons.data.handlers.AuditedEntityHandler;
import com.portlandwebworks.commons.data.handlers.BiDirectionalAssociationHandler;
import com.portlandwebworks.commons.data.handlers.EntityModificationTracker;
import com.portlandwebworks.commons.domain.Identifiable;
import com.portlandwebworks.commons.reflection.ClassUtils;
import com.portlandwebworks.commons.reflection.FieldUtils;
import com.portlandwebworks.commons.reflection.NodeType;
import com.portlandwebworks.commons.security.ISecurityProvider;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.ObjectUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;

import java.io.Serializable;
import java.lang.reflect.*;
import java.util.*;


/**
 * This software is the property of Portland Webworks.
 *
 * Possible mappings to handle:
 * - Bean field to entity field
 *
 *
 * @author Ronnie Northup
 */
public class EntityBuilder implements IEntityBuilder, InitializingBean, ApplicationContextAware
{
	private Logger logger = LoggerFactory.getLogger(EntityBuilder.class);

	@Autowired
	IEntityManager entityManager;

	Map<String, IFieldValueTransformer> transformers = new HashMap<String, IFieldValueTransformer>();

	List<IEntityHandler> entityHandlers = new ArrayList<IEntityHandler>();
	
	private ApplicationContext applicationContext;


	public EntityBuilder()
	{

	}

	public Map<String, IFieldValueTransformer> getTransformers()
	{
		return transformers;
	}

	public void setTransformers(Map<String, IFieldValueTransformer> transformers)
	{
		this.transformers = transformers;
	}

	public void setEntityHandlers(IEntityHandler[] entityHandlers)
	{
		this.entityHandlers.addAll(Arrays.asList(entityHandlers));
	}

	public void afterPropertiesSet() throws Exception
	{
		ISecurityProvider securityProvider = applicationContext.getBean(ISecurityProvider.class);

		// This handler is necessary to support basic functionality for audited entities.
		AuditedEntityHandler auditedEntityHandler = new AuditedEntityHandler();
		auditedEntityHandler.setSecurityProvider(securityProvider);
		entityHandlers.add(auditedEntityHandler);

		// This handler is necessary to manage bi-directional relationships between entities.
		entityHandlers.add(new BiDirectionalAssociationHandler());

		// Add a reference to this object to each handler.
		for (IEntityHandler entityHandler : entityHandlers) {
			entityHandler.setEntityManager(entityManager);
		}
	}

	public void setApplicationContext(ApplicationContext applicationContext) throws BeansException
	{
		this.applicationContext = applicationContext;
	}

	public Identifiable createEntity(Object bean, IEntityHandler... handlers)
	{
		return createEntity(bean, null, handlers);
	}

	public Identifiable createEntity(Object bean, String context, IEntityHandler... handlers)
	{
		// Convert the bean data provided into a graph of maps and collections of maps.
		Map valueMap = FieldUtils.mapBeanData(bean);

		// Combine the global handlers assigned to this builder with the builders passed via method arguments.
		List<IEntityHandler> combined = new ArrayList<IEntityHandler>();
		combined.addAll(entityHandlers);
		combined.addAll(Arrays.asList(handlers));

		Identifiable entity = _buildEntity(null, valueMap, context, null, (IEntityHandler[])combined.toArray(new IEntityHandler[combined.size()]));

		for (IEntityHandler handler : combined) {
			handler.entityCreated(entity);
		}
		
		return entity;
	}
	
	public void updateEntity(Identifiable entity, Object bean, IEntityHandler... handlers)
	{
		updateEntity(entity, bean,  null, handlers);
	}


	public void updateEntity(Identifiable entity, Object bean, String context, IEntityHandler... handlers)
	{
		// Convert the bean data provided into a graph of maps and collections of maps.
		Map valueMap = FieldUtils.mapBeanData(bean);

		// Combine the global handlers assigned to this builder with the builders passed via method arguments.
		// Combine the global handlers assigned to this builder with the builders passed via method arguments.
		List<IEntityHandler> combined = new ArrayList<IEntityHandler>();
		combined.addAll(entityHandlers);
		combined.addAll(Arrays.asList(handlers));
		EntityModificationTracker modificationTracker = new EntityModificationTracker();
		combined.add(modificationTracker);

		_buildEntity(entity, valueMap, context, null, (IEntityHandler[])combined.toArray(new IEntityHandler[combined.size()]));

		if (modificationTracker.isModificationDetected()) {
			for (IEntityHandler handler : combined) {
				handler.entityUpdated(entity);
			}
		}
	}

	private Identifiable _buildEntity(Identifiable entity, Map valueMap, final String context, Class<? extends Identifiable> mappedEntityType, final IEntityHandler... handlers)
	{
		// FieldUtils.mapBeanData() planted this critical value under this key value.
		Class beanClass = (Class)valueMap.get("class");

		// Get the entity class from the bean's entity mapping annoatation
		EntityMapping entityMapping = (EntityMapping)beanClass.getAnnotation(EntityMapping.class);
		if (entityMapping == null) {
			throw new RuntimeException(String.format("No entity mapping available for class %s, can't convert bean data to entity", beanClass.getName()));
		}

		// Get the entity type that the bean maps too and complain if it can't be determined.  Typically, this entity mapping
		// is defined on the bean in an @EntityMapping annotation.  However, this can be overridden in the parent bean by
		// putting a mappedClass value in the @FieldMapping annotation.
		Class<? extends Identifiable> mappedEntityClass;
		if (mappedEntityType != null) {
			logger.debug("Using entity mapping override from field level: {}", mappedEntityType.getName());
			mappedEntityClass = mappedEntityType;
		}
		else {
			mappedEntityClass = ClassUtils.getClassForNameAndContract(Identifiable.class, entityMapping.value());
		}

		// Get the field on the bean that is the entity's ID and get its value
		Field idField = getIdFieldForMappedBean(beanClass);
		
		// If no entity was passed to the build method, need to create it now.
		if (entity == null) {

			// Check the bean map for an entity identifier value.
			Serializable id = null;
			if (idField != null) {
				id = (Serializable)valueMap.get(idField.getName());
			}
			
			if (id == null) {
				logger.debug("Creating new instance of {}", mappedEntityClass.getName());
				
				entity = newEntity(mappedEntityClass);
			}
			else {
				logger.debug("Loading existing instance of {} with ID {}", mappedEntityClass.getName(), id);

				entity = loadEntity(mappedEntityClass, id);

				for (IEntityHandler handler : handlers) {
					handler.entityLoaded(entity);
				}
			}
		}

		// Iterate through all the field mappings on the
		Field[] mappedBeanFields = FieldUtils.getFieldsWithAnnotation(beanClass, FieldMapping.class);
		for (Field mappedField : mappedBeanFields) {

			final FieldMapping beanMapping = mappedField.getAnnotation(FieldMapping.class);
			if (idField != null && idField.equals(mappedField)) {
				// The ID field on the entity does not need to be processed.
				continue;
			}
			
			if (beanMapping.readOnly()) {
				logger.debug("Skipping read-only field {}", mappedField.getName());
				continue;
			}
			
			if (StringUtils.isNotEmpty(beanMapping.groups()[0])) {
				if (! ArrayUtils.contains(beanMapping.groups(), context)) {
					logger.debug("Skipping update of field {}, does not match context {} ", mappedField.getName(), context);
					continue;
				}
			}

			// Determine the path to the entity field that this field is mapped to.
			final String pathToMappedField = StringUtils.isBlank(beanMapping.value())? mappedField.getName(): beanMapping.value();

			final Object valueToApply = valueMap.get(mappedField.getName());
			
			if (valueToApply == null && beanMapping.ignoreNull()) {
				logger.debug("Ignoring null value for field {}", mappedField.getName());
				continue;
			}

			final Stack entities = new Stack();
			entities.push(entity);
			final Class[] componentType = new Class[1];

			FieldUtils.traversePathInClassGraph(entity.getClass(), pathToMappedField, new FieldUtils.FieldTypeVisitor()
			{
				public boolean visit(String path, Class<?> elementType, NodeType nodeType)
				{
					Object peek = entities.peek();

					switch (nodeType) {
						case ValueType: {
							if (beanMapping.identifier()) {
								// The value to apply is an identifier for an entity that should be loaded, not a value
								// to apply to an entity.
								List<Identifiable> currNestedEntities = (List<Identifiable>)peek;

								// Value to apply is a list of IDs for entities.  Load them all by ID and assign to the
								// collection at the top of the stack.
								Collection<Serializable> ids = (Collection<Serializable>)valueToApply;
								List<Identifiable> newNestedEntities = null;
								if (CollectionUtils.isNotEmpty(ids)) {
									newNestedEntities = entityManager.findAllByPredicate(
											componentType[0],
											Predicate.isIn("id", ids)
									);

									// Tell the handlers all about the load
									for (IEntityHandler handler : handlers) {
										handler.entitiesLoaded(newNestedEntities);
									}
								}
								else {
									// No nested entities to assign.
									newNestedEntities = Collections.emptyList();
								}

								// Get a list of the entities that will be added and removed to/from the entity's collection.
								List<Identifiable> removedEntities = ListUtils.complement(currNestedEntities, newNestedEntities);
								List<Identifiable> addedEntities = ListUtils.complement(newNestedEntities, currNestedEntities);

								// Only do the following things if the nested entities actually changed.
								if (removedEntities.size() > 0 || addedEntities.size() > 0) {

									// Get a reference to the parent entity for the handler api.  Do this by looking at
									// the stack entry one below the top.
									if (entities.size() < 2) {
										throw new RuntimeException("Invalid entity stack state: size < 2");
									}

									Object parent = entities.get(entities.size() - 2);
									ClassUtils.assertIsInstanceOf(parent, Identifiable.class, "Parent object is not identifiable");

									if (removedEntities.size() > 0) {
										// Tell the handlers about the removed entities.
										for (IEntityHandler handler : handlers) {
											handler.entitiesUnassignedFromParent((Identifiable)parent, path, removedEntities);
										}
									}

									// Clear list before assigning new members.
									currNestedEntities.clear();
									currNestedEntities.addAll(newNestedEntities);

									if (addedEntities.size() > 0) {
										// Tell the handlers about the new assignments.
										for (IEntityHandler handler : handlers) {
											handler.entitiesAssignedToParent((Identifiable)parent, path, addedEntities);
										}
									}
								}
							}
							else {
								// This field is a value mapped directly onto a field of the entity.
								Field entityField = FieldUtils.getField(peek.getClass(), path);

								// Make a modifiable copy of finally declared closure value.
								Object finalValue = valueToApply;

								// Check the field mapping annotation for a transformation, and apply if it there is one
								if (!beanMapping.transformer().equals("")) {
									IFieldValueTransformer transformer = transformers.get(beanMapping.transformer());
									if (transformer == null) {
										logger.warn("Transformer not available: " + beanMapping.transformer());

										// Transformer is not registered.  A class name might have been used as the transformer
										// field on the annotation.  Try to instantiate it.
										Class<? extends IFieldValueTransformer> transformerType = ClassUtils.getClassForNameAndContract(
												IFieldValueTransformer.class, beanMapping.transformer());

										transformer = ClassUtils.newInstanceOfClass(transformerType);
									}
									finalValue = transformer.transformFieldValue(valueToApply);
								}

								ClassUtils.assertIsInstanceOf(peek, Identifiable.class,
										"Can't assign value type to non-identifiable parent of type %s");

								Object before = FieldUtils.getFieldValue(peek, entityField);
								if (!ObjectUtils.equals(before, finalValue)) {

									FieldUtils.setFieldValue(peek, entityField, finalValue);

									for (IEntityHandler handler : handlers) {
										handler.entityFieldValueApplied((Identifiable)peek, path, before, finalValue);
									}
								}
							}
							break;
						}
						case Object: {
							// Need to determine whether the object being visited is an intermediary object or the target
							// entity to which the value field should be applied.  This looks like a pretty bad hack, but
							// I'm having a tough time poking holes in the logic.
							boolean shouldTraverseField = true;
							if (beanMapping.identifier()) {
								int depth = entities.size();
								int dotCount = StringUtils.countMatches(pathToMappedField, ".");
								shouldTraverseField = (depth < dotCount);
							}

							if (!shouldTraverseField) {
								Identifiable currValue = (Identifiable)FieldUtils.getFieldValue(peek, path);

								// Object visited is the target entity and the value to be applied is an identifier that
								// indicates which entity should be loaded and assigned to the previously visited entity
								if (valueToApply == null) {
									// Null identifier specified.  Null out the nested entity.
									if (currValue != null) {
										FieldUtils.setFieldValue(peek, path, null);

										for (IEntityHandler handler : handlers) {
											handler.entityUnassignedFromParent((Identifiable)peek, path, currValue);
										}
									}
								}
								else {
									// Load the entity by ID and assign to the current entity.
									Serializable id = (Serializable)valueToApply;
									// Only do this stuff if the nested entity actually changed.
									if (currValue == null || !currValue.getId().equals(id)) {
										Identifiable newValue = (Identifiable)entityManager.findById(elementType, id);
										FieldUtils.setFieldValue(peek, path, newValue);

										for (IEntityHandler handler : handlers) {
											handler.entityAssignedToParent((Identifiable)peek, path, newValue);
										}
									}
								}
								// We are not traversing beyond this relationship.
								return false;
							}
							else {
								if (valueToApply == null) {
									logger.debug("Null input value for nested entity.");
									return false;
								}
								
								// I don't remember why I put this validation in here :-P
								if (!(Identifiable.class.isAssignableFrom(peek.getClass()))) {
									throw new RuntimeException("Can't assign entity to parent of type " + peek.getClass());
								}

								EntityMapping nestedMapping = null;

								// This will hold the value of an override for the entity class, defined at the field level.
								Class<? extends Identifiable> entityClassOverride = null;
								if (StringUtils.isNotBlank(beanMapping.mappedClass())) {
									entityClassOverride = ClassUtils.getClassForNameAndContract(Identifiable.class, beanMapping.mappedClass());
								}

								// Look at the first item in the collection of bean inputs to determine information about the
								// entity association defined in the @EntityMapping assigned to the bean's class definition
								Field idMappingField = null;
								if (Map.class.isAssignableFrom(valueToApply.getClass())) {
									Class beanClass = (Class)((Map)valueToApply).get("class");
									idMappingField = getIdFieldForMappedBean(beanClass);
									nestedMapping = (EntityMapping)beanClass.getAnnotation(EntityMapping.class);
								}

								Serializable beanId = (Serializable)((Map)valueToApply).get(idMappingField.getName());

								// If the mapping is not an identifier, it is setting field values on a one-to-one nested entity.
								// Use the entity's accessor method to get a reference to it.  If none is found, create one.
								boolean isEntityMatch = false;
								Identifiable nestedEntity = (Identifiable)FieldUtils.getFieldValue(peek, path);
								if (nestedEntity != null) {
									if (beanId != null) {
										Serializable entityId = nestedEntity.getId();
										isEntityMatch = entityId != null && entityId.equals(beanId);
									}
									else {
										isEntityMatch = false;
									}
								}
								if (isEntityMatch) {
									// The bean data is for an update to the entity currently assigned to this field.
									_buildEntity(nestedEntity, (Map)valueToApply, context, entityClassOverride, handlers);

									if (DaoUtils.isFieldLazyLoaded(peek.getClass(), path)) {
										// If the underlying ORM had to load the entity from data source as a result of
										// the field get call, notify handlers.
										for (IEntityHandler handler : handlers) {
											handler.entityLoaded(nestedEntity);
										}
									}
								}
								else {
									if (beanId == null) {
										// The input bean has no identifier so it must be a new record.  Build a new one
										// Only create the nested entity if the bean mapping says to do so.
										if (beanMapping.autoCreate()) {
											nestedEntity = _buildEntity(null, (Map)valueToApply, context, entityClassOverride, handlers);

											for (IEntityHandler handler : handlers) {
												handler.entityCreated(nestedEntity);
											}
										}
									}
									else {
										// A bean ID that does not match the currently assigned entity's ID.  Fetch it and assign
										// it to the parent entity.
										Class<? extends Identifiable> mappedEntityClass;
										if (entityClassOverride != null) {
											mappedEntityClass = entityClassOverride;
										}
										else {
											mappedEntityClass = ClassUtils.getClassForNameAndContract(Identifiable.class, nestedMapping.value());
										}
										try {
											nestedEntity = entityManager.findById(mappedEntityClass, beanId);
											_buildEntity(nestedEntity, (Map)valueToApply, context, mappedEntityClass, handlers);
										}
										catch (DataNotFoundException e) {
											throw new RuntimeException(String.format("Could not find entity of type %s and id %s", mappedEntityClass.getName(), beanId));
										}
									}

									// Associate the nested entity with the root entity.
									FieldUtils.setFieldValue(peek, path, nestedEntity);

									for (IEntityHandler handler : handlers) {
										handler.entityAssignedToParent((Identifiable)peek, path, nestedEntity);
									}
								}
							}
							break;
						}
						case Collection: {
							if (valueToApply != null) {
								ClassUtils.assertIsInstanceOf(valueToApply, Collection.class,
										"Can't apply non-collection values to collection of nested entities: %s", path);
							}

							Field collectionField = FieldUtils.getField(peek.getClass(), path);

							// Get a reference to the nested collection on the target entity.  This will be the collection
							// that we add or remove entities to/from base on input beans provided.
							Collection<Identifiable> nestedEntities = (Collection<Identifiable>)FieldUtils.getFieldValue(peek, path);
							if (nestedEntities == null) {								
								// Null collection found on target entity.  Create a new collection for the entity based 
								// on the field's concrete type
								Class collectionType = collectionField.getType();
								nestedEntities = (Collection<Identifiable>)ClassUtils.newInstanceOfClass(collectionType);
								FieldUtils.setFieldValue(peek, path, nestedEntities);
							}

							EntityMapping nestedMapping = null;

							// This will hold the value of an override for the entity class, defined at the field level.
							Class<? extends Identifiable> entityClassOverride = null;
							if (StringUtils.isNotBlank(beanMapping.mappedClass())) {
								entityClassOverride = ClassUtils.getClassForNameAndContract(Identifiable.class, beanMapping.mappedClass());

							}

							// Look at the first item in the collection of bean inputs to determine information about the
							// entity association defined in the @EntityMapping assigned to the bean's class definition
							Field idMappingField = null;
							Collection valuesToApply = (Collection)valueToApply;
							if (CollectionUtils.isNotEmpty(valuesToApply)) {
								Object first = valuesToApply.iterator().next();
								if (Map.class.isAssignableFrom(first.getClass())) {
									Map sample = (Map)valuesToApply.iterator().next();
									Class beanClass = (Class)sample.get("class");
									idMappingField = getIdFieldForMappedBean(beanClass);
									nestedMapping = (EntityMapping)beanClass.getAnnotation(EntityMapping.class);
								}
							}

							// Build or retrieve a list of entities from the list of bean values by recursively calling
							// this method with each map of bean values in the source list.
							//TODO: an optimization would be to gather IDs for all previously persisted entities, and to
							//TODO: load them all at once.  Would require some significant code changes though.
							if (nestedMapping != null || entityClassOverride != null) {
								ArrayList<Identifiable> addedEntities = new ArrayList<Identifiable>();
								ArrayList<Identifiable> removedEntities = new ArrayList<Identifiable>();
								Set<Serializable> idSet = new HashSet<Serializable>();
								for (Object nestedBean : valuesToApply) {
									Map nestedMap = (Map)nestedBean;
									Serializable beanId = (Serializable)nestedMap.get(idMappingField.getName());
									if (beanId == null) {
										// The input bean has no identifier so it must be a new record.  Build a new one
										// and add it to the list of added entities.
										Identifiable nestedEntity = _buildEntity(null, (Map)nestedBean, context, entityClassOverride, handlers);
										addedEntities.add(nestedEntity);
									}
									else {
										// Input bean has an identifier.  Save it for later determining which nested 
										// entities were removed from the parent due to omission from the source data.
										idSet.add(beanId);

										// Look for a nested entity matching by ID in the target entity's collection, and
										// if one is found update the existing nested entity.
										boolean foundMatch = false;
										for (Identifiable nestedEntity : nestedEntities) {
											if (nestedEntity.getId().equals(beanId)) {
												_buildEntity(nestedEntity, nestedMap, context, entityClassOverride, handlers);
												foundMatch = true;
												break;
											}
										}

										// If no match was found in the target entity's collection, this is an addition.  
										// Look this bean's corresponding entity up and add it to the target entity's
										// collection,
										if (! foundMatch) {
											Class<? extends Identifiable> mappedEntityClass;
											if (entityClassOverride != null) {
												mappedEntityClass = entityClassOverride;
											}
											else {
												mappedEntityClass = ClassUtils.getClassForNameAndContract(Identifiable.class, nestedMapping.value());
											}

											try {
												Identifiable addedEntity = entityManager.findById(mappedEntityClass, beanId);
												_buildEntity(addedEntity, nestedMap, context, mappedEntityClass, handlers);
												addedEntities.add(addedEntity);
											}
											catch (DataNotFoundException e) {
												throw new RuntimeException(String.format("Could not find entity of type %s and id %s", mappedEntityClass.getName(), beanId));
											}
										}
									}
								}

								// Now that we have iterated over all the nested entities in the source data and gathered
								// the set of identifiers, iterate through the parent entity's collection and "remove"
								// all nested entities that were not in the source data.
								for (Identifiable nestedEntity : nestedEntities) {
									if (!idSet.contains(nestedEntity.getId())) {
										removedEntities.add(nestedEntity);
									}
								}

								// Handle additions.
								if (addedEntities.size() > 0) {
									nestedEntities.addAll(addedEntities);
									for (IEntityHandler handler : handlers) {
										handler.entitiesAssignedToParent((Identifiable)peek, path, addedEntities);
									}
								}

								// Handle omissions.
								if (removedEntities.size() > 0) {
									nestedEntities.removeAll(removedEntities);
									for (IEntityHandler handler : handlers) {
										handler.entitiesUnassignedFromParent((Identifiable)peek, path, removedEntities);
									}
								}
							}
							else {
								if (! valuesToApply.isEmpty()) {
									// The collection of values to apply was non-empty, but no entity mapping information could be
									// found for populating the nested entities.  This means we are working with a collection of
									// value types, presumably entity IDs, not nested form beans.  Push the collection onto the
									// stack, remember the collection's component type, and continue traversing this entity path
									// so that the builder can load each nested entity by ID and assign it to the collection.
									entities.push(nestedEntities);
									componentType[0] = elementType;
								}
								else if (! nestedEntities.isEmpty()) {
									// No entity mapping could be obtained for the nested elements because the collection of
									// values to apply was empty.  Doesn't really matter what the nested entity type is because
									// all we need to do is clear the list and let the ORM do the nested entity removal.
									List<Identifiable> removedEntities = new ArrayList<Identifiable>(nestedEntities);
									nestedEntities.clear();
									for (IEntityHandler handler : handlers) {
										handler.entitiesUnassignedFromParent((Identifiable)peek, path, removedEntities);
									}
								}
							}

							break;
						}
					}
					return true;
				}
			});
		}

		return entity;

	}

	private Field getIdFieldForMappedBean(Class beanClass)
	{
		Field[] mappedFields = FieldUtils.getFieldsWithAnnotation(beanClass, FieldMapping.class);
		for (Field mappedBeanField : mappedFields) {
			FieldMapping mapping = mappedBeanField.getAnnotation(FieldMapping.class);

			// Identifier assigned to a field directly on the entity (no dot traversal)
			if (mapping.identifier() && mapping.value().indexOf(".") == -1) {
				return mappedBeanField;
			}
		}
		return null;
	}

	private Identifiable loadEntity(Class entityClass, Serializable id)
	{
		Identifiable entity;// Do an entity lookup for the entity by ID
		try {
			entity = (Identifiable)entityManager.findById(entityClass, id);
		}
		catch (DataNotFoundException e) {
			throw new RuntimeException(String.format("Entity not found: %s(id=%s)", entityClass.getName(), id));
		}
		catch (IllegalArgumentException e) {
			throw new RuntimeException(String.format("Invalid ID value %s", id));
		}
		return entity;
	}

	private Identifiable newEntity(Class entityClass)
	{
		Identifiable entity;
		try {
			entity = (Identifiable)entityClass.newInstance();
		}
		catch (InstantiationException e) {
			throw new RuntimeException("Unable to instantiate mapped entity: " + entityClass.getName(), e);
		}
		catch (IllegalAccessException e) {
			throw new RuntimeException("Illegal access to constructor for mapped entity: " + entityClass.getName(), e);
		}
		return entity;
	}
}

