package com.portlandwebworks.commons.data.handlers;

import com.portlandwebworks.commons.domain.Identifiable;
import com.portlandwebworks.commons.reflection.ClassUtils;
import com.portlandwebworks.commons.reflection.FieldUtils;
import org.hibernate.loader.entity.CollectionElementLoader;

import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.OneToOne;
import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.util.Collection;
import java.util.List;

/**
 * This software is the property of Portland Webworks.
 *
 * This handler addresses the issue of using the IEntityBuilder on entities that have bi-directional relationships.  if
 * a handler method indicates that an entity has been assigned to a "parent", the entity's association to the parent will
 * also be set.  Conversely, if and entity is unassigned from a parent, its association to that parent will be undone.
 *
 * The presence of a bi-directional relationship is determined by evaluating whether the entity has any field of the
 * same type as the parent entity, and whether that field has a JPA annotation on it to manage that relationship.
 *
 * Do not use this handler directly or add it to an IEntityHandler instance.  It is considered a "built-in" handler that
 * is always present, as it provides necessary functionality.
 *
 * @author Ronnie Northup
 */
public class BiDirectionalAssociationHandler extends DefaultEntityHandler
{

	@Override
	public void entityAssignedToParent(Identifiable parent, String path, Identifiable entity)
	{
		Field bidiField = _getBidirectionalAssignmentField(parent, entity);
		if (bidiField != null) {
			if (Collection.class.isAssignableFrom(bidiField.getType())) {
				Collection collection = (Collection)FieldUtils.getFieldValue(entity, bidiField);
				if (collection == null) {
					collection = (Collection)ClassUtils.newInstanceOfClass(bidiField.getType());
					FieldUtils.setFieldValue(entity, bidiField, collection);
				}
				collection.add(parent);
			}
			else {
				FieldUtils.setFieldValue(entity, bidiField, parent);
			}
		}
	}

	@Override
	public void entitiesAssignedToParent(Identifiable parent, String path, List<Identifiable> entities)
	{
		Identifiable first = entities.get(0);
		Field bidiField = _getBidirectionalAssignmentField(parent, first);
		if (bidiField != null) {
			for (Identifiable entity : entities) {
				if (Collection.class.isAssignableFrom(bidiField.getType())) {
					Collection collection = (Collection)FieldUtils.getFieldValue(entity, bidiField);
					if (collection == null) {
						collection = (Collection)ClassUtils.newInstanceOfClass(bidiField.getType());
						FieldUtils.setFieldValue(entity, bidiField, collection);
					}
					collection.add(parent);
				}
				else {
					FieldUtils.setFieldValue(entity, bidiField, parent);
				}
			}
		}
	}

	@Override
	public void entityUnassignedFromParent(Identifiable parent, String path, Identifiable entity)
	{
		Field bidiField = _getBidirectionalAssignmentField(parent, entity);
		if (bidiField != null) {
			if (Collection.class.isAssignableFrom(bidiField.getType())) {
				Collection collection = (Collection)FieldUtils.getFieldValue(entity, bidiField);
				collection.remove(parent);
			}
			else {
				FieldUtils.setFieldValue(entity, bidiField, parent);
			}
		}
	}

	@Override
	public void entitiesUnassignedFromParent(Identifiable parent, String path, List<Identifiable> entities)
	{
		Identifiable first = entities.get(0);
		Field bidiField = _getBidirectionalAssignmentField(parent, first);
		if (bidiField != null) {
			for (Identifiable entity : entities) {
				if (Collection.class.isAssignableFrom(bidiField.getType())) {
					Collection collection = (Collection)FieldUtils.getFieldValue(entity, bidiField);
					collection.remove(parent);
				}
				else {
					FieldUtils.setFieldValue(entity, bidiField, parent);
				}
			}
		}
	}

	private Field _getBidirectionalAssignmentField(Identifiable parent, Identifiable entity)
	{
		Class parentClass = parent.getClass();
		Field[] entityFields = FieldUtils.getFields(entity.getClass());
		for (Field entityField : entityFields) {
			for (Annotation annotation : entityField.getAnnotations()) {
				if (annotation instanceof ManyToOne) {
					if (entityField.getType().equals(parentClass)) {
						return entityField;
					}
				}
				else if (annotation instanceof OneToOne) {
					if (entityField.getType().equals(parentClass)) {
						return entityField;
					}
				}
				else if (annotation instanceof OneToMany) {
					if (FieldUtils.getComponentTypeForField(entityField).equals(parentClass)) {
						return entityField;
					}
				}
			}
		}
		return null;
	}
}
