/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.domain;

import com.portlandwebworks.commons.dao.DefaultOrderBy;

import javax.persistence.*;

/**
 * This software is the property of Portland Webworks.
 *
 * This subclass of EntityBase provides additional columns applicable to using this entity as a lookup
 * value or as a list of options in a multi-option input field.
 *
 * By default, lookup value queries order their results by displayOrder.  You can override default sorting by adding
 * a @DefaultOrderBy annotation to your subclass field with a priority value less than 100.
 *
 * @Author rnorthup
 */
@MappedSuperclass
public class LookupValue extends Identifiable<Integer> implements Comparable
{
	@Id
	@Column(name = "id", precision = 10, scale = 0, insertable = false, updatable = false)
	private Integer id;

	@Override
	public Integer getId()
	{
		return id;
	}

	@Override
	public void setId(Integer id)
	{
		this.id = id;
	}

	@Column(name = "display_name", length = 200, nullable = false)
	private String displayName;

	@Column(name = "display_order", precision = 5, scale = 0)
	@DefaultOrderBy(priority = 100)
	private int displayOrder;

	@Column(name = "is_enabled", nullable = false)
	private boolean enabled;

	/**
	 * If this lookup value is capable of having a parent, this returns the type of that parent, or
	 * null if parents are not supported.
	 *
	 * @return
	 */
	public String getParentType()
	{
		return null;
	}

	/**
	 * Returns the parent lookup value, or null if hierarchy is not supported for this lookup type, or
	 * null if no parent has been set.
	 * @return
	 */
	public LookupValue getParent()
	{
		return null;
	}

	/**
	 *
	 * @param parent
	 */
	public void setParent(LookupValue parent)
	{
	}


	/**
	 * The name that should appear in a user interface.
	 * @return
	 */
	public String getDisplayName()
	{
		return displayName;
	}

	public void setDisplayName(String displayName)
	{
		this.displayName = displayName;
	}

	/**
	 * Provides a mechanism for ordering entities of this type for scenarios where lexicographic ordering
	 * is not appropriate.
	 *
	 * @return
	 */
	public int getDisplayOrder()
	{
		return displayOrder;
	}

	public void setDisplayOrder(int displayOrder)
	{
		this.displayOrder = displayOrder;
	}

	/**
	 * Indicates whether this lookup value should be displayed to the general user.
	 *
	 * Of course, the actual meaning of this field can be application specific.
	 *
	 * @return
	 */
	public boolean isEnabled()
	{
		return enabled;
	}

	public void setEnabled(boolean enabled)
	{
		this.enabled = enabled;
	}

	public int compareTo(Object o)
	{
		LookupValue that = (LookupValue)o;
		return this.displayOrder == that.displayOrder? 0:
				this.displayOrder < that.displayOrder? -1: 1;
	}
}