/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.domain;

import com.portlandwebworks.commons.dao.DefaultOrderBy;
import org.hibernate.annotations.Type;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.userdetails.UserDetails;

import javax.persistence.*;
import java.util.*;

/**
 * This software is the property of Portland Webworks.
 *
 * General purpose entity to support end user access to an application.
 *
 * This entity is configured to allow subclassing using a separate table that will be joined for select operations.
 *
 * The default ordering behavior of a user query is to order by lastName, firstName ascending.  You can apply your own
 * defaults by adding @DefaultOrderBy annotations to your subclass fields with priority values less than 100.
 *
 * @author Ronnie Northup
 */
@Entity
@Table(name = "user_")
@Inheritance(strategy = InheritanceType.SINGLE_TABLE)
@DiscriminatorColumn(name = "type", discriminatorType = DiscriminatorType.CHAR)
@DiscriminatorValue("U")
public class User<R extends Role> extends EntityBase implements UserDetails
{
	@Column(name = "username", length = 30, nullable = false, unique = true)
	private String username;

	@Column(name = "first_name", length = 30, nullable = false)
	@DefaultOrderBy(priority = 110)
	private String firstName;

	@Column(name = "last_name", length = 30, nullable = false)
	@DefaultOrderBy(priority = 100)
	private String lastName;

	@Column(name = "middle_initial")
	private Character middleInitial;

	@Column(name = "suffix", length = 10)
	private String suffix;

	@Column(name = "is_enabled", nullable = false)
	private boolean enabled;

	@Column(name = "last_login_date")
	@Type(type = "com.portlandwebworks.commons.hibernate.DateTimeType")
	private Date lastLoginDate;

	@Column(name = "email", length = 120)
	private String email;

	@Column(name = "creating_user_id", updatable = false)
	private Long creatingUserId;

	@Column(name = "created_date", updatable = false)
	@Type(type = "com.portlandwebworks.commons.hibernate.DateTimeType")
	private Date createdDate;

	@Column(name = "last_modifying_user_id")
	private Long lastModifyingUserId;

	@Column(name = "last_modified_date")
	@Type(type = "com.portlandwebworks.commons.hibernate.DateTimeType")
	private Date lastModifiedDate;

	@Column(name = "password_hash", length = 40)
	private String password;

	/**
	 * Maintains a space delimited list of the last X passwords used, hashed and base64 encoded.  Each time a new
	 * non-temp password is set for this user, that password is prepended to the front of this list, and the last
	 * one is removed.
	 */
	@Column(name = "password_history", length = 1000)
	private String passwordHistory;

	@Column(name = "last_pwd_change_date")
	@Type(type = "com.portlandwebworks.commons.hibernate.DateTimeType")
	private Date lastPasswordChangeDate;

	/**
	 * Certain password reset strategies require
	 */
	@Column(name = "password_reset_token", length = 100)
	private String passwordResetToken;

	@Column(name = "password_reset_date")
	private Date passwordResetDate;

	/**
	 * This field is not stored in the database.  It should be set by the service tier based on whatever business logic
	 * exists around password enforcement.
	 */
	@Transient
	private boolean passwordExpired;

	@ManyToMany 
	@JoinTable(name = "user_role",
			joinColumns = { @JoinColumn(name = "user_id") },
			inverseJoinColumns = @JoinColumn(name = "role_id"))
	private List<R> roles = new ArrayList<R>();

	@Transient
	private ArrayList<GrantedAuthority> authorities;


	public String getUsername()
	{
		return username;
	}

	public boolean isAccountNonExpired()
	{
		return enabled;
	}

	public boolean isAccountNonLocked()
	{
		return enabled;
	}

	public boolean isCredentialsNonExpired()
	{
		return !passwordExpired;
	}

	public void setUsername(String username)
	{
		this.username = username;
	}

	public String getFirstName()
	{
		return firstName;
	}

	public void setFirstName(String firstName)
	{
		this.firstName = firstName;
	}

	public String getFirstAndLastName()
	{
		return getFirstName() + " " + getLastName();
	}

	public String getLastAndFirstName()
	{
		return getLastName() + ", " + getFirstName();
	}

	public String getLastName()
	{
		return lastName;
	}

	public void setLastName(String lastName)
	{
		this.lastName = lastName;
	}

	public Character getMiddleInitial()
	{
		return middleInitial;
	}

	public void setMiddleInitial(Character middleInitial)
	{
		this.middleInitial = middleInitial;
	}

	public String getSuffix()
	{
		return suffix;
	}

	public void setSuffix(String suffix)
	{
		this.suffix = suffix;
	}

	public boolean isEnabled()
	{
		return enabled;
	}

	public void setEnabled(boolean enabled)
	{
		this.enabled = enabled;
	}

	public Date getLastLoginDate()
	{
		return lastLoginDate;
	}

	public void setLastLoginDate(Date lastLoginDate)
	{
		this.lastLoginDate = lastLoginDate;
	}

	public String getEmail()
	{
		return email;
	}

	public void setEmail(String email)
	{
		this.email = email;
	}

	public Long getCreatingUserId()
	{
		return creatingUserId;
	}

	public void setCreatingUserId(Long creatingUserId)
	{
		this.creatingUserId = creatingUserId;
	}

	public Date getCreatedDate()
	{
		return createdDate;
	}

	public void setCreatedDate(Date createdDate)
	{
		this.createdDate = createdDate;
	}

	public Long getLastModifyingUserId()
	{
		return lastModifyingUserId;
	}

	public void setLastModifyingUserId(Long lastModifyingUserId)
	{
		this.lastModifyingUserId = lastModifyingUserId;
	}

	public Date getLastModifiedDate()
	{
		return lastModifiedDate;
	}

	public void setLastModifiedDate(Date lastModifiedDate)
	{
		this.lastModifiedDate = lastModifiedDate;
	}

	public String getPassword()
	{
		return password;
	}

	public void setPassword(String password)
	{
		this.password = password;
	}

	public String getPasswordHistory()
	{
		return passwordHistory;
	}

	public void setPasswordHistory(String passwordHistory)
	{
		this.passwordHistory = passwordHistory;
	}

	public List<R> getRoles()
	{
		return roles;
	}

	public void setRoles(List<R> roles)
	{
		this.roles = roles;
	}

	/**
	 * Returns true if the user has been assigned the role whose code matches the string representation of the enum
	 * provided.
	 *
	 * @param roleCode  The enumeration whose string value will be tested.
	 * @return
	 */
	public boolean isInRole(Enum roleCode)
	{
		for (Role role : roles) {
			if (role.getCode().equals(roleCode.toString())) {
				return true;
			}
		}
		return false;
	}

	public Collection<GrantedAuthority> getAuthorities()
	{
		if (authorities == null) {
			authorities = new ArrayList<GrantedAuthority>();
		}
		return authorities;
	}

	public void setAuthorities(Collection<GrantedAuthority> authorities)
	{
		if (authorities instanceof ArrayList) {
			this.authorities = (ArrayList<GrantedAuthority>)authorities;
		}
		else {
			this.authorities = new ArrayList<GrantedAuthority>(authorities);
		}
	}

	public Date getLastPasswordChangeDate()
	{
		return lastPasswordChangeDate;
	}

	public void setLastPasswordChangeDate(Date lastPasswordChangeDate)
	{
		this.lastPasswordChangeDate = lastPasswordChangeDate;
	}

	public boolean isPasswordExpired()
	{
		return passwordExpired;
	}

	public void setPasswordExpired(boolean passwordExpired)
	{
		this.passwordExpired = passwordExpired;
	}

	public String getPasswordResetToken()
	{
		return passwordResetToken;
	}

	public void setPasswordResetToken(String passwordResetToken)
	{
		this.passwordResetToken = passwordResetToken;
	}

	public Date getPasswordResetDate()
	{
		return passwordResetDate;
	}

	public void setPasswordResetDate(Date passwordResetDate)
	{
		this.passwordResetDate = passwordResetDate;
	}
}
