/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.email;

import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;
import java.util.*;

/**
 * This software is the property of Portland Webworks.
 *
 * Waits for the completion of the issuance of one or more related emails by the EmailSender, and reports on the success
 * and failure of each individual email.
 *
 * @author Ronnie Northup
 */
public class EmailResult
{
	/**
	 * The bucket of all emails in the batch that have not yet been sent.
	 */
	private final Set<MimeMessage> pending = new HashSet<MimeMessage>();

	/**
	 * The bucket of all strings from which a valid email address could not be parsed.
	 */
	private List<String> badAddresses = new ArrayList<String>();

	/**
	 * The bucket of all email failures
	 */
	private Map<MimeMessage, Exception> failures = new HashMap<MimeMessage, Exception>();

	/**
	 * The bucket of all email successes
	 */
	private List<MimeMessage> successes = new ArrayList<MimeMessage>();

	/**
	 * The bucket of all emails that were determined to be duplicates of previously queued emails.
	 */
	private List<MimeMessage> duplicates = new ArrayList<MimeMessage>();

	/**
	 * Wait for the asynchronous task of sending email(s) to complete so that the success/failure of their issuance can
	 * be determined.
	 *
	 */
	public void waitFor()
	{
		waitFor(Long.MAX_VALUE);
	}

	/**
	 * Wait for the asynchronous task of sending email(s) to complete so that the success/failure of their issuance can
	 * be determined.  Will time out after the specified number of milliseconds.
	 *
	 * Timeout can be determined by calling #hasAnyPending() after this method returns and checking for a true value.
	 *
	 * Note that this method may not time out in the specified number of milliseconds if an I/O dependency such as an
	 * underlying socket is not configured to time out in less time.
	 *
	 * @param millis    The number of milliseconds to wait before timing out.
	 */
	public void waitFor(long millis)
	{
		// Pending collection will be empty once all messages have been processed.
		if (! pending.isEmpty()) {
			synchronized (pending) {
				// Always do a second check first thing in a synchronized block.
				if (! pending.isEmpty()) {
					try {
						//  
						pending.wait(millis);
					}
					catch (InterruptedException e) {
					}
				}
			}
		}
	}

	private void markComplete()
	{
		// Any process that called waitFor() is using the pending collection as a mutex.
		synchronized (pending) {
			pending.notify();
		}
	}

	/**
	 * Register this recipient as part of the current batch.
	 * 
	 * @param message
	 */
	void addPending(MimeMessage message)
	{
		pending.add(message);
	}

	/**
	 * Returns true if any of the emails from this batch have not yet be sent to the mail server.
	 * @return
	 */
	public boolean hasAnyPending()
	{
		return pending.size() > 0;
	}

	/**
	 * An email address could not be parsed from this text.
	 *
	 * @param emailAddress
	 */
	void addBadAddress(String emailAddress)
	{
		badAddresses.add(emailAddress);

		if (! hasAnyPending()) {
			markComplete();
		}
	}

	/**
	 * Email sent to this address failed.
	 *
	 * @param message
	 * @param cause
	 */
	void setFailure(MimeMessage message, Exception cause)
	{
		if (! pending.remove(message)) {
			throw new RuntimeException("Email address " + message + " not in list of pending recipients");
		}

		failures.put(message, cause);

		if (! hasAnyPending()) {
			markComplete();
		}
	}

	/**
	 * If this email address is still in a pending state, consider it to be a failed delivery.
	 *
	 * @param message
	 */
	void setUnknownFailure(MimeMessage message)
	{
		if (pending.remove(message)) {
			failures.put(message, new RuntimeException("Unknown error"));
		}

		if (! hasAnyPending()) {
			markComplete();
		}
	}

	/**
	 * Reports that email sent to this address was successfully delivered.
	 *
	 * @param message
	 */
	void setSuccess(MimeMessage message)
	{
		if (! pending.remove(message)) {
			throw new RuntimeException("Message " + message + " not in list of pending recipients");
		}

		successes.add(message);

		if (! hasAnyPending()) {
			markComplete();
		}
	}

	/**
	 * The list of all email addresses that were of a format not recognized by the sender.
	 * @return
	 */
	public List<String> getBadAddresses()
	{
		return badAddresses;
	}

	/**
	 * The list of all email addresses to which an email was successfully sent, as indicated by the mail server response
	 * or due to internal error.
	 *
	 * @return
	 */
	public List<MimeMessage> getSuccesses()
	{
		return new ArrayList<MimeMessage>(successes);
	}

	/**
	 * The list of all emails that were failed to be sent, as indicated by the mail server response or due to internal
	 * error.
	 *
	 * @return
	 */
	public List<MimeMessage> getFailures()
	{
		return new ArrayList<MimeMessage>(failures.keySet());
	}

	public List<MimeMessage> getDuplicates()
	{
		return new ArrayList<MimeMessage>(duplicates);
	}

	/**
	 * Given an email address that the email sender failed to deliver to, provides the exception that was reported by
	 * the mail server, email transport layer or by the sender itself.
	 *
	 * @param message
	 * @return
	 */
	public Exception getFailureCause(MimeMessage message)
	{
		return failures.get(message);
	}

	/**
	 * This method is used internally by the EmailSender to mark an email as a duplicate of another email in the queue.
	 * This is in support of the duplicate message suppression functionality.  Marking a message as a duplicate means
	 * that the message was not sent because another message in the send queue was determined to be a match of this one.
	 *
	 * @param message
	 */
	void setDuplicate(MimeMessage message)
	{
		if (! pending.remove(message)) {
			throw new RuntimeException("Message " + message + " not in list of pending recipients");
		}

		duplicates.add(message);

		if (! hasAnyPending()) {
			markComplete();
		}
	}
}
