package com.portlandwebworks.commons.jsf.lifeCycle;

import org.apache.commons.lang.StringUtils;

import javax.faces.context.FacesContext;
import javax.servlet.ServletContext;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This software is the property of Portland Webworks.
 *
 * This implementation of PhaseCallbackSupport determines the name of a view's backing bean by parsing the text out of
 * the view path and joining it together in camel-case format with the file extension stripped off.  It leverages a
 * regular expression to determine what elements of the view path make up the backing bean name.
 *
 * For example, if a view has the path /admin/userList.xhtml, the corresponding backing bean could be adminUserList.
 * This is what would be derived by the built-in regular expression.
 *
 * If your view files are nested under a common directory, you could use a regular expression like this:
 *
 * (?:/(?!view/components)(?!components)([\w\d\-_]+))
 *
 * to strip off the /view/components text that prefixes all views located under the directory /view/components.
 *
 * If you like to prefix or suffix your backing bean names, this is not the implementation for you.
 * 
 *
 * @author Ronnie Northup
 */
public class RegexPhaseCallbackSupport extends PhaseCallbackSupport
{
	/**
	 * The regular expression that will be used to parse View IDs.
	 */
	private Pattern beanNameRegex;

	/**
	 * By default, the bean name can be derived from the JSF view ID by joining the names of the parent directories and
	 * the file name of the view and camel casing the entire thing.  For example, /admin/user/list.xhtml would have a
	 * backing bean named adminUserList.  Similarly, /admin/userList.xhtml would have a backing bean named adminUserList.
	 */
	public static final String DEFAULT_BEAN_NAME_PATTERN = "(?:/([\\w\\d\\-_]+))";

	
	@Override
	public String getBeanNameForViewPath(FacesContext facesContext, String viewPath)
	{
		if (beanNameRegex == null) {
			ServletContext servletContext = (ServletContext)facesContext.getExternalContext().getContext();
			String pattern = servletContext.getInitParameter("RegexPhaseCallbackSupport.beanNameRegex");
			if (StringUtils.isBlank(pattern)) {
				pattern = DEFAULT_BEAN_NAME_PATTERN;
			}
			beanNameRegex = Pattern.compile(pattern);
		}

		String beanName = null;

		Matcher matcher = beanNameRegex.matcher(viewPath);
		if (matcher.find()) {
			switch (matcher.groupCount()) {
				case 0:
					// No match.
					return null;
				case 1:
					// One capture was specified.  Skip the entire matching group and just return the single capture.
					beanName = matcher.group(1);

					// If the single capture is meant to be repeated over the entire view ID, do repeated searches
					// across the string.
					boolean another = matcher.find();
					if (another) {
						StringBuilder sb = new StringBuilder(beanName);
						while (another) {
							String group = matcher.group(1);
							if (StringUtils.isNotBlank(group)) {
								sb.append(StringUtils.capitalize(matcher.group(1)));
							}
							another = matcher.find();
						}
						beanName = sb.toString();
					}

					break;
				default:
					// Multiple groupings were specified.  Join them together, camel casing the entire thing
					StringBuilder sb = new StringBuilder();
					for (int i = 1; i < matcher.groupCount(); i++) {
						if (sb.length() == 0) {
							sb.append(matcher.group(i));
						}
						else {
							sb.append(StringUtils.capitalize(matcher.group(i)));
						}
					}
					beanName = sb.toString();
			}
		}
		return beanName;
	}
}
