package com.portlandwebworks.commons.pdf;

import com.lowagie.text.BadElementException;
import com.lowagie.text.Image;
import org.apache.commons.io.IOUtils;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.core.io.ClassPathResource;
import org.w3c.dom.Element;
import org.xhtmlrenderer.extend.FSImage;
import org.xhtmlrenderer.extend.ReplacedElement;
import org.xhtmlrenderer.extend.ReplacedElementFactory;
import org.xhtmlrenderer.extend.UserAgentCallback;
import org.xhtmlrenderer.layout.LayoutContext;
import org.xhtmlrenderer.layout.SharedContext;
import org.xhtmlrenderer.pdf.ITextFSImage;
import org.xhtmlrenderer.pdf.ITextImageElement;
import org.xhtmlrenderer.render.BlockBox;
import org.xhtmlrenderer.resource.ImageResource;
import org.xhtmlrenderer.simple.extend.FormSubmissionListener;

import java.io.IOException;
import java.io.InputStream;
import java.util.LinkedHashMap;

/**
 * Allows Flying Saucer PDF renderer to load an HTML file's resources from the classpath.
 *
 * @author John Gordon
 */
public class ClasspathReplacedElementFactory implements ReplacedElementFactory {

	private static final Logger logger = LogManager.getLogger(ClasspathReplacedElementFactory.class);

	private UserAgentCallback agentCallback;

	private String baseURL = "";
	private LinkedHashMap<String, ImageResource> imageCache = new LinkedHashMap<String, ImageResource>();
	private SharedContext sharedContext;


	public ClasspathReplacedElementFactory(SharedContext sharedContext) {

		this.sharedContext = sharedContext;
	}

	public ReplacedElement createReplacedElement(LayoutContext c, BlockBox box, UserAgentCallback uac, int cssWidth, int cssHeight) {

		Element e = box.getElement();
		if (e == null) {
			return null;
		}

		String nodeName = e.getNodeName();
		String path = e.getAttribute("src");

		if (nodeName.equals("img")) {
			logger.debug("Fetching image " + path);
			FSImage fsImage = getImageResource(path).getImage();
			if (fsImage != null) {
				if (cssWidth != -1 || cssHeight != -1) {
					fsImage.scale(cssWidth, cssHeight);
				}
				return new ITextImageElement(fsImage);
			}

		}

		return null;
	}

	public ImageResource getImageResource(String uri) {
		ImageResource resource;
		resource = imageCache.get(uri);
		if (resource == null) {
			InputStream is = resolveAndOpenStream(uri);
			if (is != null) {
				try {
					logger.debug("Found image " + uri);
					Image image = Image.getInstance(IOUtils.toByteArray(is));
					scaleToOutputResolution(image);
					resource = new ImageResource(new ITextFSImage(image));

					imageCache.put(uri, resource);
				} catch (IOException e) {
					logger.error("Can't read image file; unexpected problem for URI '" + uri + "'", e);
				} catch (BadElementException e) {
					logger.error("Can't read image file; unexpected problem for URI '" + uri + "'", e);
				} finally {
					try {
						is.close();
					} catch (IOException e) {
						// ignore
					}
				}
			}
		}
		if (resource == null) {
			logger.warn("Couldn't find image " + uri);
			resource = new ImageResource(null);
		}
		return resource;
	}


	protected InputStream resolveAndOpenStream(String uri) {

		ClassPathResource resource = new ClassPathResource(uri);

		try {
			return resource.getInputStream();
		} catch (IOException e) {
			throw new RuntimeException(e.getMessage(), e);
		}
	}


	private void scaleToOutputResolution(Image image) {
		float factor = sharedContext.getDotsPerPixel();
		image.scaleAbsolute(image.getPlainWidth() * factor, image.getPlainHeight() * factor);
	}


	public void reset() {
	}

	public void remove(Element e) {
	}

	public void setFormSubmissionListener(FormSubmissionListener listener) {
	}


}
