package com.portlandwebworks.commons.reflection;

import com.portlandwebworks.commons.domain.Identifiable;

import java.lang.reflect.Modifier;
import java.util.*;

/**
 * This software is the property of Portland Webworks.
 *
 * @author Ronnie Northup
 */
public class ClassUtils
{
	public static <T> Class<? extends T> getClassForNameAndContract(Class<T> contract, String name) throws IllegalArgumentException
	{
		try {
			Class temp = Class.forName(name);

			if (! contract.isAssignableFrom(temp)) {
				throw new IllegalArgumentException(String.format("Class %s does not implement contract %s", name, contract.getName()));
			}

			return (Class<? extends T>)temp;
		}
		catch (ClassNotFoundException e) {
			throw new IllegalArgumentException("Could not find class by name " + name);
		}
	}

	public static <T> T newInstanceOfClass(Class<T> clazz)
	{
		try {
			if (Modifier.isAbstract(clazz.getModifiers())) {

				if (List.class.isAssignableFrom(clazz)) {
					return (T)new ArrayList<Identifiable>();
				}
				else if (Map.class.isAssignableFrom(clazz)) {
					return (T)new HashMap();
				}
				else if (Set.class.isAssignableFrom(clazz)) {
					return (T)new HashSet<Identifiable>();
				}
				else {
					throw new RuntimeException("Don't know of a concrete subclass of abstract collection " + clazz.getName());
				}
			}
			else {
				return clazz.newInstance();
			}
		}
		catch (InstantiationException e) {
			throw new RuntimeException("Unable to instantiate class using default constructor: " + clazz.getName());
		}
		catch (IllegalAccessException e) {
			throw new RuntimeException("No public default constructor available for class: " + clazz.getName());
		}
	}

	public static void assertIsInstanceOf(Object obj, Class clazz)
	{
		assertIsInstanceOf(obj, clazz, "Object is not an instance of or subclass of %s", clazz.getName());
	}

	public static void assertIsInstanceOf(Object obj, Class clazz, String message, Object... params)
	{
		if (! clazz.isAssignableFrom(obj.getClass())) {
			String formattedMessage;
			if (message.indexOf("%s") != -1 && params.length == 0) {
				formattedMessage = String.format(message, clazz.getName());
			}
			else {
				formattedMessage = String.format(message, params);
			}
			throw new RuntimeException(formattedMessage);
		}
	}

	/**
	 * Does a rudimentary type conversion of an object to another type.
	 *
	 * @param value
	 * @param type
	 * @return
	 */
	public static Object convertToType(Object value, Class type)
	{
		if (value == null) {
			return null;
		}

		Class sourceType = value.getClass();

		if (type.equals(sourceType) || type.isAssignableFrom(sourceType)) {
			return value;
		}

		if (type.equals(String.class)) {
			return value.toString();
		}
		else if (Boolean.class.isAssignableFrom(type)) {
			return Boolean.valueOf(value.toString());
		}
		else if (Number.class.isAssignableFrom(type)) {
			Double dbl = null;
			try {
				dbl = Double.valueOf(value.toString());
			}
			catch (NumberFormatException e) {
				throw new RuntimeException(String.format("Failed to convert value %s to type ", value, type));
			}

			if (Integer.class.isAssignableFrom(type)) {
				return dbl.intValue();
			}
			else if (Long.class.isAssignableFrom(type)) {
				return dbl.longValue();
			}
			else if (Float.class.isAssignableFrom(type)) {
				return dbl.floatValue();
			}
			else if (Double.class.isAssignableFrom(type)) {
				return dbl;
			}
		}
		else if (Character.class.isAssignableFrom(type)) {
			return type.toString().charAt(0);
		}
		return value;
	}
}
