package com.portlandwebworks.commons.service;

import com.portlandwebworks.commons.validation.CustomValidationException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.web.context.WebApplicationContext;
import org.springframework.web.context.support.WebApplicationContextUtils;

import javax.servlet.*;
import javax.validation.ConstraintViolation;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Map;

/**
 * This software is the property of Portland Webworks.
 *
 * @author Ronnie Northup
 */
public class ServiceOperationResultFilter implements Filter
{
	private Logger logger = LoggerFactory.getLogger(ServiceOperationResultFilter.class);

	protected WebApplicationContext springContext;

	private JsonObjectMapper jsonMapper;

	public void init(FilterConfig filterConfig) throws ServletException
	{
		ServletContext servletContext = filterConfig.getServletContext();

		springContext = WebApplicationContextUtils.getRequiredWebApplicationContext(servletContext);

		jsonMapper = springContext.getBean(JsonObjectMapper.class);
		if (jsonMapper == null) {
			throw new ServletException("Could not locate a JsonObjectMapper in the Spring context of this web application");
		}
	}


	public void doFilter(ServletRequest request, ServletResponse response, FilterChain filterChain) throws IOException, ServletException
	{
		try {
			filterChain.doFilter(request, response);
		}
		catch (Exception ex) {

			Throwable e = ex.getCause();

			if (e != null) {

				logger.error("Caught unhandled exception: " + e.getMessage(), e);

				ServiceOperationResult<Void> result = new ServiceOperationResult<Void>(
						null,
						false,
						e.getMessage());
				
				if (e.getMessage() == null) {
					result.setMessage("Unknown error occurred");
				}

				if (e instanceof CustomValidationException) {
					CustomValidationException cve = (CustomValidationException)e;

					Map<String, String> errors = new HashMap<String, String>();

					for (Object violation : cve.getViolations()) {
						ConstraintViolation cv = (ConstraintViolation)violation;
						errors.put(cv.getPropertyPath().toString(), cv.getMessage());
					}

					result.setErrors(errors);
				}

				// Print this data into memory so that we can find out how large it is and send back a proper content
				// length response header
				ByteArrayOutputStream baos = new ByteArrayOutputStream();
				jsonMapper.writeValue(new OutputStreamWriter(baos), result);
				byte[] bytes = baos.toByteArray();

				response.setContentType("application/json");
				response.setContentLength(bytes.length);

				final PrintWriter writer = response.getWriter();
				writer.write(new String(bytes, "UTF-8"));
				writer.close();
			}
			else if (ex instanceof ServletException) {
				throw (ServletException)e;
			}
			else {
				throw new ServletException("A system error has occurred", ex);
			}
		}
	}

	public void destroy()
	{
	}
}
