package com.portlandwebworks.commons.spring.security;


import com.portlandwebworks.commons.servlet.FilterBean;
import com.portlandwebworks.commons.servlet.PatternMatchingFilter;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.context.SecurityContextHolder;

import javax.servlet.*;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;


/**
 * This software is the property of Portland Webworks.
 *
 * This filter can be used to redirect users to a change password page when they have he ROLE_CHANGE_PASSWORD authority,
 * which is added to the user's spring authentication object during the login process if the user needs to change their
 * password to access the application.
 *
 * This filter can be added to web.xml--however, it is recommended that it is added to your spring security config like
 * this:
 *
 * <code>

 	<beans:bean id="changePasswordFilter" class="com.portlandwebworks.commons.spring.security.PasswordChangeEnforcingFilter">
		<beans:property name="passwordChangeUrl" value="/passwordChange.html"/>
		<beans:property name="urlPatterns">
			<beans:list>
				<beans:value>/</beans:value>
				<beans:value>/index.html</beans:value>
				<beans:value>/service/**</beans:value>
			</beans:list>
		</beans:property>
	</beans:bean>

    <http>
        <custom-filter ref="changePasswordFilter" before="FILTER_SECURITY_INTERCEPTOR"/>
    </http>

  </code>
 *
 * @author Ronnie Northup
 */
public class PasswordChangeEnforcingFilter extends PatternMatchingFilter implements InitializingBean
{
	private Logger logger = LoggerFactory.getLogger(PasswordChangeEnforcingFilter.class);
	
	private String passwordChangeUrl;
	

	/**
	 * Tells this filter what URL to ignore for enforcing password change, so that the user can access the password
	 * change page. This URL will not restrict access to users with the password change role.
	 *
	 * @param passwordChangeUrl
	 */
	public void setPasswordChangeUrl(String passwordChangeUrl)
	{
		this.passwordChangeUrl = passwordChangeUrl;
	}


	public void afterPropertiesSet() throws Exception
	{
		if (StringUtils.isBlank(this.passwordChangeUrl)) {
			throw new ServletException("Bad configuration for filter.  A passwordChangeUrl is required");
		}

		// Add the password change URL to the list of URLs excluded from filtering
		this.setExcludeUrlPatterns((String[])ArrayUtils.add(this.getExcludeUrlPatterns(), this.passwordChangeUrl));
	}

	@Override
	public void urlIncluded(HttpServletRequest request, HttpServletResponse response) throws IOException
	{
		Authentication auth = SecurityContextHolder.getContext().getAuthentication();
		if (auth != null) {
			logger.debug("Checking authentication '{}' for ROLE_CHANGE_PASSWORD authority", auth.getName());

			for (GrantedAuthority authority : auth.getAuthorities()) {
				if (authority.getAuthority().equals("ROLE_CHANGE_PASSWORD")) {
					String redirect = passwordChangeUrl;
					if (passwordChangeUrl.startsWith("/")) {
						redirect = request.getContextPath() + passwordChangeUrl;
					}

					logger.info("Authenticated user has ROLE_CHANGE_PASSWORD authority.  Redirecting to {}", redirect);

					response.sendRedirect(redirect);

					return;
				}
			}
		}
	}

	@Override
	public void urlExcluded(HttpServletRequest request, HttpServletResponse response) throws IOException
	{
		// Do nothing
	}

}
