/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.spring.security;

import com.portlandwebworks.commons.security.AuthenticationFailedReason;
import com.portlandwebworks.commons.security.AuthenticationResult;
import com.portlandwebworks.commons.security.ISecurityProvider;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.authority.GrantedAuthorityImpl;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.core.userdetails.UserDetails;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

/**
 * Implementation of the ISecurityProvider interface which utilizes a Spring security context.
 *
 * This software is the property of Portland Webworks.
 *
 * @author Ronnie Northup
 */
public class SpringSecurityProvider implements ISecurityProvider
{
	@Autowired
	@Qualifier("org.springframework.security.authenticationManager")
	private AuthenticationManager authManager;

	public AuthenticationResult authenticate(String login, String password)
	{
		AuthenticationResult result = new AuthenticationResult();

		Authentication auth = authManager.authenticate(new UsernamePasswordAuthenticationToken(login, password));

		// Store the authentication result in spring's security context.
		SecurityContextHolder.getContext().setAuthentication(auth);

		result.setAuthenticated(auth.isAuthenticated());

		// Translate failure to authenticate into a generic enumerated value.
		if (! auth.isAuthenticated()) {
			result.setFailureReason(AuthenticationFailedReason.BadCredentials);
		}
		else {
			// Get the principal from the successful auth
			UserDetails principal = (UserDetails)auth.getPrincipal();

			// Check the principal for error conditions, and if one is found translate to generic enumerated value.
			if (! principal.isAccountNonExpired()) {
				result.setFailureReason(AuthenticationFailedReason.AccountExpired);
			}
			else if (! principal.isAccountNonLocked()) {
				result.setFailureReason(AuthenticationFailedReason.AccountLocked);
			}
			else if (! principal.isCredentialsNonExpired()) {
				result.setFailureReason(AuthenticationFailedReason.PasswordExpired);
			}
			else if (! principal.isEnabled()) {
				result.setFailureReason(AuthenticationFailedReason.AccountDisabled);
			}

			// Populate the generic auth result's principal field with the Spring principal's username.
			result.setPrincipal(principal.getUsername());

			// Pull the authorities from the Spring principal into the generic auth result's role list.
			List<String> roles = new ArrayList<String>(principal.getAuthorities().size());
			for (GrantedAuthority authority : principal.getAuthorities()) {
				roles.add(authority.getAuthority());
			}
			result.setRoles(roles);
		}

		return result;
	}

	public void changeAuthentication(Object principal, String password, String... roles)
	{
		List<GrantedAuthority> authorities = new ArrayList<GrantedAuthority>(roles.length);
		for (String role : roles) {
			authorities.add(new GrantedAuthorityImpl(role));
		}

		UsernamePasswordAuthenticationToken token =
				new UsernamePasswordAuthenticationToken(principal, password, authorities);

		SecurityContextHolder.getContext().setAuthentication(token);
	}

	public void logout()
	{
		SecurityContextHolder.clearContext();
	}

	private Authentication getAuthentication() throws SecurityException
	{
		SecurityContext context = SecurityContextHolder.getContext();
		if (context != null) {
			Authentication authentication = context.getAuthentication();
			if (authentication != null) {
				return authentication;
			}
		}
		return null;
	}

	public boolean isAuthenticated()
	{
		Authentication authentication = getAuthentication();

		return authentication != null && authentication.isAuthenticated() && !"anonymousUser".equals(authentication.getName());
	}

	public Object getPrincipal() throws SecurityException
	{
		Authentication auth = getAuthentication();
		if (auth == null) {
			throw new SecurityException("Not authenticated");
		}

		Object principal = auth.getPrincipal();
		if (principal == null) {
			throw new SecurityException("No principal found in security context");
		}

		return principal;
	}

	public List<String> getRoles()
	{
		Authentication auth = getAuthentication();
		if (auth == null) {
			throw new SecurityException("Not authenticated");
		}

		Collection<? extends GrantedAuthority> authorities = auth.getAuthorities();
		List<String> roles = new ArrayList<String>(authorities.size());
		for (GrantedAuthority authority : authorities) {
			roles.add(authority.getAuthority());
		}
		return roles;
	}
	
	public boolean hasRole(Object role)
	{
		Authentication auth = getAuthentication();
		if (auth == null) {
			throw new SecurityException("Not authenticated");
		}

		Collection<? extends GrantedAuthority> authorities = auth.getAuthorities();
		for (GrantedAuthority authority : authorities) {
			if (authority.getAuthority().equals(role.toString())) {
				return true;
			}
		}
		return false;
	}
	
	public void assertHasRole(Object role) throws AccessDeniedException
	{
		if (! hasRole(role)) {
			throw new AccessDeniedException("Access denied.  User does not have role '" + role + "'");
		}
	}


	public boolean hasAnyRole(Object... roles)
	{
		Authentication auth = getAuthentication();
		if (auth == null) {
			throw new SecurityException("Not authenticated");
		}

		Collection<? extends GrantedAuthority> authorities = auth.getAuthorities();
		for (Object role : roles) {
			for (GrantedAuthority authority : authorities) {
				if (authority.getAuthority().equals(role.toString())) {
					return true;
				}
			}
		}
		return false;
	}
	
	public void assertHasAnyRole(Object... roles) throws AccessDeniedException
	{
		if (! hasAnyRole(roles)) {
			throw new AccessDeniedException("Access denied.  User does not have one of the roles '" + Arrays.toString(roles) + "'");
		}
	}
}
