package com.portlandwebworks.commons.test.liquibase;

import liquibase.Liquibase;
import liquibase.changelog.ChangeSet;
import liquibase.database.DatabaseConnection;
import liquibase.database.jvm.HsqlConnection;
import liquibase.database.jvm.JdbcConnection;
import liquibase.exception.LiquibaseException;
import liquibase.exception.RollbackImpossibleException;
import liquibase.resource.ClassLoaderResourceAccessor;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.ApplicationContext;
import org.springframework.test.context.TestContext;
import org.springframework.test.context.TestExecutionListener;

import javax.sql.DataSource;
import java.lang.reflect.Method;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.List;

/**
 * This software is the property of Portland Webworks.
 *
 * @author Ronnie Northup
 */
public class LiquibaseTestExecutionListener implements TestExecutionListener
{
	private Logger logger = LoggerFactory.getLogger(LiquibaseTestExecutionListener.class);
	

	public void beforeTestClass(TestContext testContext) throws Exception
	{
		ChangeSets changeSets = testContext.getTestClass().getAnnotation(ChangeSets.class);
		if (changeSets != null) {
			DatabaseConnection connection = null;
			try {
				connection = buildDatabaseConnection(changeSets, testContext.getApplicationContext());

				for (int i = 0; i < changeSets.value().length; i++) {
	
					String changeSetFile = changeSets.value()[i];
	
					Liquibase liquibase = new Liquibase(changeSetFile, new ClassLoaderResourceAccessor(), connection);
	
					if (i == 0 && !(connection instanceof HsqlConnection)) {
						// If about to apply the first changeset, and the test class/method annotation specified drop all,
						// do it but only do so if we're NOT running on HSQL.  There is no reason to drop all on an in
						// memory database because no db objects exist on startup.  More importantly, I'm getting a
						// strange liquibase exception doing drop all on a schema that has a view in it.
						if (changeSets.dropAll()) {
							liquibase.dropAll();
						}
					}
	
					liquibase.update(changeSets.contexts());
	
					if (i == changeSets.value().length - 1) {
						// If we just ran the last class-level changeset in the list, apply a tag for convenient rollbacks
						String rollbackTag = "rollback." + testContext.getTestClass().getName();
	
						testContext.setAttribute("rollbackTag", rollbackTag);
	
						liquibase.tag(rollbackTag);
					}
				}
			}
			finally {
				if (connection != null) {
					connection.close();
				}
			}
		}
	}

	public void prepareTestInstance(TestContext testContext) throws Exception
	{
	}

	public void beforeTestMethod(TestContext testContext) throws Exception
	{
		ChangeSets changeSets = testContext.getTestMethod().getAnnotation(ChangeSets.class);
		if (changeSets != null) {
			DatabaseConnection connection = null;
			try {
				connection = buildDatabaseConnection(changeSets, testContext.getApplicationContext());

				for (int i = 0; i < changeSets.value().length; i++) {
	
					String changeSetFile = changeSets.value()[i];
	
					Liquibase liquibase = new Liquibase(changeSetFile, new ClassLoaderResourceAccessor(), connection);
	
					liquibase.update(changeSets.contexts());
				}
			}
			finally {
				if (connection != null) {
					connection.close();
				}
			}
		}
	}

	public void afterTestMethod(TestContext testContext) throws Exception
	{
		String rollbackTag = (String)testContext.getAttribute("rollbackTag");

		ChangeSets changeSets = testContext.getTestMethod().getAnnotation(ChangeSets.class);
		if (changeSets != null) {
			DatabaseConnection connection = null;
			try {
				connection = buildDatabaseConnection(changeSets, testContext.getApplicationContext());

				for (int i = 0; i < changeSets.value().length; i++) {

					String changeSetFile = changeSets.value()[i];

					Liquibase liquibase = new Liquibase(changeSetFile, new ClassLoaderResourceAccessor(), connection);

					liquibase.rollback(rollbackTag, changeSets.contexts());
				}
			}
			catch (RollbackImpossibleException e) {
				throw new UnsupportedOperationException(
						"Your liquibase changeset must have a rollback element for unit testing.  If you have no " +
						"rollback steps to perform, simply add an empty <rollback></rollback> element");
			}
			finally {
				if (connection != null) {
					connection.close();
				}
			}
		}
	}

	public void afterTestClass(TestContext testContext) throws Exception
	{
	}
	
	private DatabaseConnection buildDatabaseConnection(ChangeSets changeSets, ApplicationContext context)
	{
		DataSource ds;

		if (changeSets.dataSource().equals("")) {
			// Load by name
			ds = context.getBean(DataSource.class);
		}
		else {
			// Load by type
			ds = context.getBean(changeSets.dataSource(), DataSource.class);
		}
		
		DatabaseConnection connection = null;

		try {
			// Liquibase forces you to use a different connection class depending on whether you're using HSQL or a non
			// HSQL database.  To determine whether we're running on HSQL or otherwise, try running an HQL-specific
			// query and if that throws an exception, we're non-HSQL.
			Connection conn = ds.getConnection();
			Statement stmt = conn.createStatement();
			try {
				stmt.execute("SELECT 1 FROM INFORMATION_SCHEMA.SYSTEM_USERS");
				connection = new HsqlConnection(conn);
			}
			catch (SQLException e) {
				connection = new JdbcConnection(conn);
			}
		}
		catch (SQLException e) {
			throw new RuntimeException("Could not establish database connection using data source " + ds);
		}
		
		return connection;
	}
}
